/*
 * Copyright (c) 2008 Lukas Mejdrech
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup net
 * @{
 */

/** @file
 */

#include <async.h>
#include <stdio.h>

#include <ipc/ipc.h>
#include <ipc/services.h>
//#include <sys/mman.h>

#include "../err.h"
#include "../messages.h"
#include "../modules.h"

#include "../structures/packet/packet.h"
#include "../structures/packet/packet_client.h"

#include "device.h"
#include "netif.h"

#define IPC_GET_DEVICE( call )		( device_id_t ) IPC_GET_ARG1( * call )
#define IPC_GET_PACKET( call )		( packet_id_t ) IPC_GET_ARG2( * call )
#define IPC_GET_IRQ( call )		( int ) IPC_GET_ARG2( * call )
#define IPC_GET_IO( call )		( int ) IPC_GET_ARG3( * call )
#define IPC_GET_PHONE( call )		( int ) IPC_GET_ARG5( * call )

extern netif_globals_t netif_globals;

extern int	initialize( void );
extern int	probe_auto_message( void );
extern int	probe_message( device_id_t device_id, int irq, int io );
extern int	send_message( device_id_t device_id, packet_t packet );
extern int	start_message( device_id_t device_id );
extern int	stop_message( device_id_t device_id );

DEVICE_MAP_IMPLEMENT( device_map, device_t )

int	module_message( ipc_callid_t callid, ipc_call_t * call, ipc_call_t * answer, int * answer_count );
int	netif_start_module( async_client_conn_t client_connection );
int	register_message( device_id_t device_id, int phone );
int	get_device_stats( device_id_t device_id, device_stats_ref * stats );

int find_device( device_id_t device_id, device_ref * device ){
	* device = device_map_find( & netif_globals.device_map, device_id );
	if( ! * device ) return ENOENT;
	if(( ** device ).state == NETIF_NULL ) return EPERM;
	return EOK;
}

int get_device_stats( device_id_t device_id, device_stats_ref * stats ){
	ERROR_DECLARE;

	device_ref	device;

	if( ! stats ) return EINVAL;
	ERROR_PROPAGATE( find_device( device_id, & device ));
	* stats = & device->stats;
	return EOK;
}

void null_device_stats( device_stats_ref stats ){
	memset( stats, 0, sizeof( device_stats_t ));
/*	stats->rx_packets = 0;
	stats->tx_packets = 0;
	stats->rx_bytes = 0;
	stats->tx_bytes = 0;
	stats->rx_errors = 0;
	stats->tx_errors = 0;
	stats->rx_dropped = 0;
	stats->tx_dropped = 0;
	stats->multicast = 0;
	stats->collisions = 0;
	stats->rx_length_errors = 0;
	stats->rx_over_errors = 0;
	stats->rx_crc_errors = 0;
	stats->rx_frame_errors = 0;
	stats->rx_fifo_errors = 0;
	stats->rx_missed_errors = 0;
	stats->tx_aborted_errors = 0;
	stats->tx_carrier_errors = 0;
	stats->tx_fifo_errors = 0;
	stats->tx_heartbeat_errors = 0;
	stats->tx_window_errors = 0;
	stats->rx_compressed = 0;
	stats->tx_compressed = 0;
*/
}

int register_message( device_id_t device_id, int phone ){
	ERROR_DECLARE;

	device_ref	device;

	ERROR_PROPAGATE( find_device( device_id, & device ));
	if( device->nil_phone ) return ELIMIT;
	device->nil_phone = phone;
	return EOK;
}

int netif_message( ipc_callid_t callid, ipc_call_t * call, ipc_call_t * answer, int * answer_count ){
	ERROR_DECLARE;

	size_t			length;
	device_stats_ref	stats;
	packet_t		packet;

//	printf( "\nNETIF message %d", method );
	* answer_count = 0;
	switch( IPC_GET_METHOD( * call )){
		case IPC_M_PHONE_HUNGUP:
			return EOK;
		case NET_NETIF_PROBE_AUTO:
			return probe_auto_message();
		case NET_NETIF_PROBE:
			return probe_message( IPC_GET_DEVICE( call ), IPC_GET_IRQ( call ), IPC_GET_IO( call ));
		case IPC_M_CONNECT_TO_ME:
			return register_message( IPC_GET_DEVICE( call ), IPC_GET_PHONE( call ));
		case NET_NETIF_SEND:
			ERROR_PROPAGATE( packet_translate( netif_globals.networking_phone, & packet, IPC_GET_PACKET( call )));
			return send_message( IPC_GET_DEVICE( call ), packet );
		case NET_NETIF_START:
			return start_message( IPC_GET_DEVICE( call ));
		case NET_NETIF_STATS:
			ERROR_PROPAGATE( ipc_data_read_receive( & callid, & length ));
			if( length < sizeof( device_stats_t )) return EOVERFLOW;
			ERROR_PROPAGATE( get_device_stats( IPC_GET_DEVICE( call ), & stats ));
			return ipc_data_read_finalize( callid, stats, sizeof( device_stats_t ));
		case NET_NETIF_STOP:
			return stop_message( IPC_GET_DEVICE( call ));
	}
	return ENOTSUP;
}

int netif_start_module( async_client_conn_t client_connection ){
	ERROR_DECLARE;

	async_set_client_connection( client_connection );
	netif_globals.networking_phone = connect_to_service( SERVICE_NETWORKING );
	device_map_initialize( & netif_globals.device_map );
	ERROR_PROPAGATE( initialize());
	ERROR_PROPAGATE( pm_init());
	async_manager();

	return EOK;
}

/** @}
 */
